"""
    Agilicus API

    Agilicus is API-first. Modern software is controlled by other software, is open, is available for you to use the way you want, securely, simply.  The OpenAPI Specification in YAML format is available on [www](https://www.agilicus.com/www/api/agilicus-openapi.yaml) for importing to other tools.  A rendered, online viewable and usable version of this specification is available at [api](https://www.agilicus.com/api). You may try the API inline directly in the web page. To do so, first obtain an Authentication Token (the simplest way is to install the Python SDK, and then run `agilicus-cli --issuer https://MYISSUER get-token`). You will need an org-id for most calls (and can obtain from `agilicus-cli --issuer https://MYISSUER list-orgs`). The `MYISSUER` will typically be `auth.MYDOMAIN`, and you will see it as you sign-in to the administrative UI.  This API releases on Bearer-Token authentication. To obtain a valid bearer token you will need to Authenticate to an Issuer with OpenID Connect (a superset of OAUTH2).  Your \"issuer\" will look like https://auth.MYDOMAIN. For example, when you signed-up, if you said \"use my own domain name\" and assigned a CNAME of cloud.example.com, then your issuer would be https://auth.cloud.example.com.  If you selected \"use an Agilicus supplied domain name\", your issuer would look like https://auth.myorg.agilicus.cloud.  For test purposes you can use our [Python SDK](https://pypi.org/project/agilicus/) and run `agilicus-cli --issuer https://auth.MYDOMAIN get-token`.  This API may be used in any language runtime that supports OpenAPI 3.0, or, you may use our [Python SDK](https://pypi.org/project/agilicus/), our [Typescript SDK](https://www.npmjs.com/package/@agilicus/angular), or our [Golang SDK](https://git.agilicus.com/pub/sdk-go).  100% of the activities in our system our API-driven, from our web-admin, through our progressive web applications, to all internals: there is nothing that is not accessible.  For more information, see [developer resources](https://www.agilicus.com/developer).   # noqa: E501

    The version of the OpenAPI document: 2024.01.29
    Contact: dev@agilicus.com
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from agilicus_api.api_client import ApiClient, Endpoint as _Endpoint
from agilicus_api.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types
)
from agilicus_api.model.error_message import ErrorMessage
from agilicus_api.model.list_standalone_rule_trees_response import ListStandaloneRuleTreesResponse
from agilicus_api.model.list_standalone_rules_response import ListStandaloneRulesResponse
from agilicus_api.model.list_standalone_ruleset_bundles_response import ListStandaloneRulesetBundlesResponse
from agilicus_api.model.list_standalone_ruleset_labels_response import ListStandaloneRulesetLabelsResponse
from agilicus_api.model.list_standalone_rulesets_response import ListStandaloneRulesetsResponse
from agilicus_api.model.standalone_rule import StandaloneRule
from agilicus_api.model.standalone_rule_tree import StandaloneRuleTree
from agilicus_api.model.standalone_ruleset import StandaloneRuleset
from agilicus_api.model.standalone_ruleset_bundle import StandaloneRulesetBundle
from agilicus_api.model.standalone_ruleset_label import StandaloneRulesetLabel
from agilicus_api.model.trusted_certificate_label_name import TrustedCertificateLabelName


class RulesApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        def __create_ruleset_label(
            self,
            standalone_ruleset_label,
            **kwargs
        ):
            """Creates a StandaloneRulesetLabel  # noqa: E501

            Creates a StandaloneRulesetLabel   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.create_ruleset_label(standalone_ruleset_label, async_req=True)
            >>> result = thread.get()

            Args:
                standalone_ruleset_label (StandaloneRulesetLabel):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                StandaloneRulesetLabel
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['standalone_ruleset_label'] = \
                standalone_ruleset_label
            return self.call_with_http_info(**kwargs)

        if self.create_ruleset_label is None:
            self.create_ruleset_label = _Endpoint(
                settings={
                    'response_type': (StandaloneRulesetLabel,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_ruleset_labels',
                    'operation_id': 'create_ruleset_label',
                    'http_method': 'POST',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'standalone_ruleset_label',
                    ],
                    'required': [
                        'standalone_ruleset_label',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'standalone_ruleset_label':
                            (StandaloneRulesetLabel,),
                    },
                    'attribute_map': {
                    },
                    'location_map': {
                        'standalone_ruleset_label': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__create_ruleset_label
            )

        def __create_standalone_rule(
            self,
            standalone_rule,
            **kwargs
        ):
            """Add a standalone rule  # noqa: E501

            Adds a new standalone rule. Rules must have unique names within an org. If the name is not unique, a 409 will be returned.   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.create_standalone_rule(standalone_rule, async_req=True)
            >>> result = thread.get()

            Args:
                standalone_rule (StandaloneRule):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                StandaloneRule
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['standalone_rule'] = \
                standalone_rule
            return self.call_with_http_info(**kwargs)

        if self.create_standalone_rule is None:
            self.create_standalone_rule = _Endpoint(
                settings={
                    'response_type': (StandaloneRule,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_rules',
                    'operation_id': 'create_standalone_rule',
                    'http_method': 'POST',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'standalone_rule',
                    ],
                    'required': [
                        'standalone_rule',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'standalone_rule':
                            (StandaloneRule,),
                    },
                    'attribute_map': {
                    },
                    'location_map': {
                        'standalone_rule': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__create_standalone_rule
            )

        def __create_standalone_rule_tree(
            self,
            standalone_rule_tree,
            **kwargs
        ):
            """Add a standalone rule tree  # noqa: E501

            Adds a new standalone rule tree. Rule trees must have unique names within an org. If the name is not unique, a 409 will be returned.   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.create_standalone_rule_tree(standalone_rule_tree, async_req=True)
            >>> result = thread.get()

            Args:
                standalone_rule_tree (StandaloneRuleTree):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                StandaloneRuleTree
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['standalone_rule_tree'] = \
                standalone_rule_tree
            return self.call_with_http_info(**kwargs)

        if self.create_standalone_rule_tree is None:
            self.create_standalone_rule_tree = _Endpoint(
                settings={
                    'response_type': (StandaloneRuleTree,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_rule_trees',
                    'operation_id': 'create_standalone_rule_tree',
                    'http_method': 'POST',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'standalone_rule_tree',
                    ],
                    'required': [
                        'standalone_rule_tree',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'standalone_rule_tree':
                            (StandaloneRuleTree,),
                    },
                    'attribute_map': {
                    },
                    'location_map': {
                        'standalone_rule_tree': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__create_standalone_rule_tree
            )

        def __create_standalone_ruleset(
            self,
            standalone_ruleset,
            **kwargs
        ):
            """Add a standalone ruleset  # noqa: E501

            Adds a new standalone ruleset. rulesets must have unique names within an org. If the name is not unique, a 409 will be returned.   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.create_standalone_ruleset(standalone_ruleset, async_req=True)
            >>> result = thread.get()

            Args:
                standalone_ruleset (StandaloneRuleset):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                StandaloneRuleset
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['standalone_ruleset'] = \
                standalone_ruleset
            return self.call_with_http_info(**kwargs)

        if self.create_standalone_ruleset is None:
            self.create_standalone_ruleset = _Endpoint(
                settings={
                    'response_type': (StandaloneRuleset,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_rulesets',
                    'operation_id': 'create_standalone_ruleset',
                    'http_method': 'POST',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'standalone_ruleset',
                    ],
                    'required': [
                        'standalone_ruleset',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'standalone_ruleset':
                            (StandaloneRuleset,),
                    },
                    'attribute_map': {
                    },
                    'location_map': {
                        'standalone_ruleset': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__create_standalone_ruleset
            )

        def __create_standalone_ruleset_bundle(
            self,
            standalone_ruleset_bundle,
            **kwargs
        ):
            """Add a standalone ruleset bundle  # noqa: E501

            Adds a new standalone ruleset bundle. rulesets must have unique names within an org. If the name is not unique, a 409 will be returned.   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.create_standalone_ruleset_bundle(standalone_ruleset_bundle, async_req=True)
            >>> result = thread.get()

            Args:
                standalone_ruleset_bundle (StandaloneRulesetBundle):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                StandaloneRulesetBundle
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['standalone_ruleset_bundle'] = \
                standalone_ruleset_bundle
            return self.call_with_http_info(**kwargs)

        if self.create_standalone_ruleset_bundle is None:
            self.create_standalone_ruleset_bundle = _Endpoint(
                settings={
                    'response_type': (StandaloneRulesetBundle,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_ruleset_bundles',
                    'operation_id': 'create_standalone_ruleset_bundle',
                    'http_method': 'POST',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'standalone_ruleset_bundle',
                    ],
                    'required': [
                        'standalone_ruleset_bundle',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'standalone_ruleset_bundle':
                            (StandaloneRulesetBundle,),
                    },
                    'attribute_map': {
                    },
                    'location_map': {
                        'standalone_ruleset_bundle': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__create_standalone_ruleset_bundle
            )

        def __delete_ruleset_label(
            self,
            label,
            **kwargs
        ):
            """Delete a StandaloneRulesetLabel  # noqa: E501

            Delete a StandaloneRulesetLabel  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.delete_ruleset_label(label, async_req=True)
            >>> result = thread.get()

            Args:
                label (TrustedCertificateLabelName): A TrustedCertificateLabelName

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                None
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['label'] = \
                label
            return self.call_with_http_info(**kwargs)

        if self.delete_ruleset_label is None:
            self.delete_ruleset_label = _Endpoint(
                settings={
                    'response_type': None,
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_ruleset_labels/{label}',
                    'operation_id': 'delete_ruleset_label',
                    'http_method': 'DELETE',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'label',
                        'org_id',
                    ],
                    'required': [
                        'label',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'label':
                            (TrustedCertificateLabelName,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'label': 'label',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'label': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__delete_ruleset_label
            )

        def __delete_standalone_rule(
            self,
            rule_id,
            **kwargs
        ):
            """Delete a standalone rule  # noqa: E501

            Delete a standalone rule  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.delete_standalone_rule(rule_id, async_req=True)
            >>> result = thread.get()

            Args:
                rule_id (str): The id of a rule

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                None
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['rule_id'] = \
                rule_id
            return self.call_with_http_info(**kwargs)

        if self.delete_standalone_rule is None:
            self.delete_standalone_rule = _Endpoint(
                settings={
                    'response_type': None,
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_rules/{rule_id}',
                    'operation_id': 'delete_standalone_rule',
                    'http_method': 'DELETE',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'rule_id',
                        'org_id',
                    ],
                    'required': [
                        'rule_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'rule_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'rule_id': 'rule_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'rule_id': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__delete_standalone_rule
            )

        def __delete_standalone_rule_tree(
            self,
            standalone_rule_tree_id,
            **kwargs
        ):
            """Delete a standalone rule tree  # noqa: E501

            Delete a standalone rule tree  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.delete_standalone_rule_tree(standalone_rule_tree_id, async_req=True)
            >>> result = thread.get()

            Args:
                standalone_rule_tree_id (str): The id of a rule tree

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                None
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['standalone_rule_tree_id'] = \
                standalone_rule_tree_id
            return self.call_with_http_info(**kwargs)

        if self.delete_standalone_rule_tree is None:
            self.delete_standalone_rule_tree = _Endpoint(
                settings={
                    'response_type': None,
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_rule_trees/{standalone_rule_tree_id}',
                    'operation_id': 'delete_standalone_rule_tree',
                    'http_method': 'DELETE',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'standalone_rule_tree_id',
                        'org_id',
                    ],
                    'required': [
                        'standalone_rule_tree_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'standalone_rule_tree_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'standalone_rule_tree_id': 'standalone_rule_tree_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'standalone_rule_tree_id': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__delete_standalone_rule_tree
            )

        def __delete_standalone_ruleset(
            self,
            standalone_ruleset_id,
            **kwargs
        ):
            """Delete a standalone ruleset  # noqa: E501

            Delete a standalone ruleset  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.delete_standalone_ruleset(standalone_ruleset_id, async_req=True)
            >>> result = thread.get()

            Args:
                standalone_ruleset_id (str): The id of a ruleset

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                None
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['standalone_ruleset_id'] = \
                standalone_ruleset_id
            return self.call_with_http_info(**kwargs)

        if self.delete_standalone_ruleset is None:
            self.delete_standalone_ruleset = _Endpoint(
                settings={
                    'response_type': None,
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_rulesets/{standalone_ruleset_id}',
                    'operation_id': 'delete_standalone_ruleset',
                    'http_method': 'DELETE',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'standalone_ruleset_id',
                        'org_id',
                    ],
                    'required': [
                        'standalone_ruleset_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'standalone_ruleset_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'standalone_ruleset_id': 'standalone_ruleset_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'standalone_ruleset_id': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__delete_standalone_ruleset
            )

        def __delete_standalone_ruleset_bundle(
            self,
            standalone_ruleset_bundle_id,
            **kwargs
        ):
            """Delete a standalone ruleset_bundle  # noqa: E501

            Delete a standalone ruleset_bundle  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.delete_standalone_ruleset_bundle(standalone_ruleset_bundle_id, async_req=True)
            >>> result = thread.get()

            Args:
                standalone_ruleset_bundle_id (str): The id of a ruleset

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                None
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['standalone_ruleset_bundle_id'] = \
                standalone_ruleset_bundle_id
            return self.call_with_http_info(**kwargs)

        if self.delete_standalone_ruleset_bundle is None:
            self.delete_standalone_ruleset_bundle = _Endpoint(
                settings={
                    'response_type': None,
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_ruleset_bundles/{standalone_ruleset_bundle_id}',
                    'operation_id': 'delete_standalone_ruleset_bundle',
                    'http_method': 'DELETE',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'standalone_ruleset_bundle_id',
                        'org_id',
                    ],
                    'required': [
                        'standalone_ruleset_bundle_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'standalone_ruleset_bundle_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'standalone_ruleset_bundle_id': 'standalone_ruleset_bundle_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'standalone_ruleset_bundle_id': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__delete_standalone_ruleset_bundle
            )

        def __get_ruleset_label(
            self,
            label,
            **kwargs
        ):
            """Get a StandaloneRulesetLabel  # noqa: E501

            Get a StandaloneRulesetLabel  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.get_ruleset_label(label, async_req=True)
            >>> result = thread.get()

            Args:
                label (TrustedCertificateLabelName): A TrustedCertificateLabelName

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                StandaloneRulesetLabel
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['label'] = \
                label
            return self.call_with_http_info(**kwargs)

        if self.get_ruleset_label is None:
            self.get_ruleset_label = _Endpoint(
                settings={
                    'response_type': (StandaloneRulesetLabel,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_ruleset_labels/{label}',
                    'operation_id': 'get_ruleset_label',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'label',
                        'org_id',
                    ],
                    'required': [
                        'label',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'label':
                            (TrustedCertificateLabelName,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'label': 'label',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'label': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__get_ruleset_label
            )

        def __get_standalone_rule(
            self,
            rule_id,
            **kwargs
        ):
            """Get a standalone rule  # noqa: E501

            Get a standalone rule  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.get_standalone_rule(rule_id, async_req=True)
            >>> result = thread.get()

            Args:
                rule_id (str): The id of a rule

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                StandaloneRule
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['rule_id'] = \
                rule_id
            return self.call_with_http_info(**kwargs)

        if self.get_standalone_rule is None:
            self.get_standalone_rule = _Endpoint(
                settings={
                    'response_type': (StandaloneRule,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_rules/{rule_id}',
                    'operation_id': 'get_standalone_rule',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'rule_id',
                        'org_id',
                    ],
                    'required': [
                        'rule_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'rule_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'rule_id': 'rule_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'rule_id': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__get_standalone_rule
            )

        def __get_standalone_rule_tree(
            self,
            standalone_rule_tree_id,
            **kwargs
        ):
            """Get a standalone rule tree  # noqa: E501

            Get a standalone rule tree  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.get_standalone_rule_tree(standalone_rule_tree_id, async_req=True)
            >>> result = thread.get()

            Args:
                standalone_rule_tree_id (str): The id of a rule tree

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                StandaloneRuleTree
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['standalone_rule_tree_id'] = \
                standalone_rule_tree_id
            return self.call_with_http_info(**kwargs)

        if self.get_standalone_rule_tree is None:
            self.get_standalone_rule_tree = _Endpoint(
                settings={
                    'response_type': (StandaloneRuleTree,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_rule_trees/{standalone_rule_tree_id}',
                    'operation_id': 'get_standalone_rule_tree',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'standalone_rule_tree_id',
                        'org_id',
                    ],
                    'required': [
                        'standalone_rule_tree_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'standalone_rule_tree_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'standalone_rule_tree_id': 'standalone_rule_tree_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'standalone_rule_tree_id': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__get_standalone_rule_tree
            )

        def __get_standalone_ruleset(
            self,
            standalone_ruleset_id,
            **kwargs
        ):
            """Get a standalone ruleset  # noqa: E501

            Get a standalone ruleset  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.get_standalone_ruleset(standalone_ruleset_id, async_req=True)
            >>> result = thread.get()

            Args:
                standalone_ruleset_id (str): The id of a ruleset

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                StandaloneRuleset
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['standalone_ruleset_id'] = \
                standalone_ruleset_id
            return self.call_with_http_info(**kwargs)

        if self.get_standalone_ruleset is None:
            self.get_standalone_ruleset = _Endpoint(
                settings={
                    'response_type': (StandaloneRuleset,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_rulesets/{standalone_ruleset_id}',
                    'operation_id': 'get_standalone_ruleset',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'standalone_ruleset_id',
                        'org_id',
                    ],
                    'required': [
                        'standalone_ruleset_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'standalone_ruleset_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'standalone_ruleset_id': 'standalone_ruleset_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'standalone_ruleset_id': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__get_standalone_ruleset
            )

        def __get_standalone_ruleset_bundle(
            self,
            standalone_ruleset_bundle_id,
            **kwargs
        ):
            """Get a standalone ruleset bundle  # noqa: E501

            Get a standalone ruleset bundle  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.get_standalone_ruleset_bundle(standalone_ruleset_bundle_id, async_req=True)
            >>> result = thread.get()

            Args:
                standalone_ruleset_bundle_id (str): The id of a ruleset

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                get_rulesets (bool): When querying a bundle, return all rulesets associated with bundle . [optional]
                standalone_rulesets_etag (str): The entity tag (etag) for a requested bundle. If the returned etag matches the requested etag, then no data is returned, along with status code 304. . [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                StandaloneRulesetBundle
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['standalone_ruleset_bundle_id'] = \
                standalone_ruleset_bundle_id
            return self.call_with_http_info(**kwargs)

        if self.get_standalone_ruleset_bundle is None:
            self.get_standalone_ruleset_bundle = _Endpoint(
                settings={
                    'response_type': (StandaloneRulesetBundle,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_ruleset_bundles/{standalone_ruleset_bundle_id}',
                    'operation_id': 'get_standalone_ruleset_bundle',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'standalone_ruleset_bundle_id',
                        'org_id',
                        'get_rulesets',
                        'standalone_rulesets_etag',
                    ],
                    'required': [
                        'standalone_ruleset_bundle_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'standalone_ruleset_bundle_id':
                            (str,),
                        'org_id':
                            (str,),
                        'get_rulesets':
                            (bool,),
                        'standalone_rulesets_etag':
                            (str,),
                    },
                    'attribute_map': {
                        'standalone_ruleset_bundle_id': 'standalone_ruleset_bundle_id',
                        'org_id': 'org_id',
                        'get_rulesets': 'get_rulesets',
                        'standalone_rulesets_etag': 'standalone_rulesets_etag',
                    },
                    'location_map': {
                        'standalone_ruleset_bundle_id': 'path',
                        'org_id': 'query',
                        'get_rulesets': 'query',
                        'standalone_rulesets_etag': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__get_standalone_ruleset_bundle
            )

        def __list_ruleset_labels(
            self,
            **kwargs
        ):
            """list StandaloneRulesetLabel  # noqa: E501

            List StandaloneRulesetLabel   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.list_ruleset_labels(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                limit (int): limit the number of rows in the response. [optional] if omitted the server will use the default value of 500
                org_id (str): Organisation Unique identifier. [optional]
                label (str): Filters based on whether or not the items in the collection have the given label. . [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ListStandaloneRulesetLabelsResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        if self.list_ruleset_labels is None:
            self.list_ruleset_labels = _Endpoint(
                settings={
                    'response_type': (ListStandaloneRulesetLabelsResponse,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_ruleset_labels',
                    'operation_id': 'list_ruleset_labels',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'limit',
                        'org_id',
                        'label',
                    ],
                    'required': [],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                        'limit',
                    ]
                },
                root_map={
                    'validations': {
                        ('limit',): {

                            'inclusive_maximum': 500,
                            'inclusive_minimum': 1,
                        },
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'limit':
                            (int,),
                        'org_id':
                            (str,),
                        'label':
                            (str,),
                    },
                    'attribute_map': {
                        'limit': 'limit',
                        'org_id': 'org_id',
                        'label': 'label',
                    },
                    'location_map': {
                        'limit': 'query',
                        'org_id': 'query',
                        'label': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__list_ruleset_labels
            )

        def __list_standalone_rule_trees(
            self,
            **kwargs
        ):
            """List all standalone rule trees  # noqa: E501

            List all standalone rule tres matching the provided query parameters. Perform keyset pagination by setting the page_at_id parameter to the id for the next page to fetch. Set it to `\"\"` to start from the beginning. Searching for rule_id will return any tree that references that rule.   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.list_standalone_rule_trees(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                limit (int): limit the number of rows in the response. [optional] if omitted the server will use the default value of 500
                name (str): Filters based on whether or not the items in the collection have the given name. . [optional]
                page_at_id (str): Pagination based query with the id as the key. To get the initial entries supply an empty string. On subsequent requests, supply the `page_at_id` field from the list response. . [optional]
                rule_tree_id (str): The id of a rule tree. [optional]
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ListStandaloneRuleTreesResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        if self.list_standalone_rule_trees is None:
            self.list_standalone_rule_trees = _Endpoint(
                settings={
                    'response_type': (ListStandaloneRuleTreesResponse,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_rule_trees',
                    'operation_id': 'list_standalone_rule_trees',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'limit',
                        'name',
                        'page_at_id',
                        'rule_tree_id',
                        'org_id',
                    ],
                    'required': [],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                        'limit',
                    ]
                },
                root_map={
                    'validations': {
                        ('limit',): {

                            'inclusive_maximum': 500,
                            'inclusive_minimum': 1,
                        },
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'limit':
                            (int,),
                        'name':
                            (str,),
                        'page_at_id':
                            (str,),
                        'rule_tree_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'limit': 'limit',
                        'name': 'name',
                        'page_at_id': 'page_at_id',
                        'rule_tree_id': 'rule_tree_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'limit': 'query',
                        'name': 'query',
                        'page_at_id': 'query',
                        'rule_tree_id': 'query',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__list_standalone_rule_trees
            )

        def __list_standalone_rules(
            self,
            **kwargs
        ):
            """List all standalone rules  # noqa: E501

            List all standalone rule matching the provided query parameters. Perform keyset pagination by setting the page_at_id parameter to the id for the next page to fetch. Set it to `\"\"` to start from the beginning.   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.list_standalone_rules(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                limit (int): limit the number of rows in the response. [optional] if omitted the server will use the default value of 500
                name (str): Filters based on whether or not the items in the collection have the given name. . [optional]
                page_at_id (str): Pagination based query with the id as the key. To get the initial entries supply an empty string. On subsequent requests, supply the `page_at_id` field from the list response. . [optional]
                rule_id (str): The id of a rule. [optional]
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ListStandaloneRulesResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        if self.list_standalone_rules is None:
            self.list_standalone_rules = _Endpoint(
                settings={
                    'response_type': (ListStandaloneRulesResponse,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_rules',
                    'operation_id': 'list_standalone_rules',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'limit',
                        'name',
                        'page_at_id',
                        'rule_id',
                        'org_id',
                    ],
                    'required': [],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                        'limit',
                    ]
                },
                root_map={
                    'validations': {
                        ('limit',): {

                            'inclusive_maximum': 500,
                            'inclusive_minimum': 1,
                        },
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'limit':
                            (int,),
                        'name':
                            (str,),
                        'page_at_id':
                            (str,),
                        'rule_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'limit': 'limit',
                        'name': 'name',
                        'page_at_id': 'page_at_id',
                        'rule_id': 'rule_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'limit': 'query',
                        'name': 'query',
                        'page_at_id': 'query',
                        'rule_id': 'query',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__list_standalone_rules
            )

        def __list_standalone_ruleset_bundles(
            self,
            **kwargs
        ):
            """List all standalone ruleset bundles  # noqa: E501

            List all standalone ruleset bundles matching the provided query parameters. Perform keyset pagination by setting the page_at_id parameter to the id for the next page to fetch. Set it to `\"\"` to start from the beginning.   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.list_standalone_ruleset_bundles(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                limit (int): limit the number of rows in the response. [optional] if omitted the server will use the default value of 500
                name (str): Filters based on whether or not the items in the collection have the given name. . [optional]
                page_at_id (str): Pagination based query with the id as the key. To get the initial entries supply an empty string. On subsequent requests, supply the `page_at_id` field from the list response. . [optional]
                org_id (str): Organisation Unique identifier. [optional]
                label (str): Filters based on whether or not the items in the collection have the given label. . [optional]
                get_rulesets (bool): When querying a bundle, return all rulesets associated with bundle . [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ListStandaloneRulesetBundlesResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        if self.list_standalone_ruleset_bundles is None:
            self.list_standalone_ruleset_bundles = _Endpoint(
                settings={
                    'response_type': (ListStandaloneRulesetBundlesResponse,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_ruleset_bundles',
                    'operation_id': 'list_standalone_ruleset_bundles',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'limit',
                        'name',
                        'page_at_id',
                        'org_id',
                        'label',
                        'get_rulesets',
                    ],
                    'required': [],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                        'limit',
                    ]
                },
                root_map={
                    'validations': {
                        ('limit',): {

                            'inclusive_maximum': 500,
                            'inclusive_minimum': 1,
                        },
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'limit':
                            (int,),
                        'name':
                            (str,),
                        'page_at_id':
                            (str,),
                        'org_id':
                            (str,),
                        'label':
                            (str,),
                        'get_rulesets':
                            (bool,),
                    },
                    'attribute_map': {
                        'limit': 'limit',
                        'name': 'name',
                        'page_at_id': 'page_at_id',
                        'org_id': 'org_id',
                        'label': 'label',
                        'get_rulesets': 'get_rulesets',
                    },
                    'location_map': {
                        'limit': 'query',
                        'name': 'query',
                        'page_at_id': 'query',
                        'org_id': 'query',
                        'label': 'query',
                        'get_rulesets': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__list_standalone_ruleset_bundles
            )

        def __list_standalone_rulesets(
            self,
            **kwargs
        ):
            """List all standalone rulesets  # noqa: E501

            List all standalone rulesets matching the provided query parameters. Perform keyset pagination by setting the page_at_id parameter to the id for the next page to fetch. Set it to `\"\"` to start from the beginning.   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.list_standalone_rulesets(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                limit (int): limit the number of rows in the response. [optional] if omitted the server will use the default value of 500
                name (str): Filters based on whether or not the items in the collection have the given name. . [optional]
                page_at_id (str): Pagination based query with the id as the key. To get the initial entries supply an empty string. On subsequent requests, supply the `page_at_id` field from the list response. . [optional]
                org_id (str): Organisation Unique identifier. [optional]
                label (str): Filters based on whether or not the items in the collection have the given label. . [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ListStandaloneRulesetsResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        if self.list_standalone_rulesets is None:
            self.list_standalone_rulesets = _Endpoint(
                settings={
                    'response_type': (ListStandaloneRulesetsResponse,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_rulesets',
                    'operation_id': 'list_standalone_rulesets',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'limit',
                        'name',
                        'page_at_id',
                        'org_id',
                        'label',
                    ],
                    'required': [],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                        'limit',
                    ]
                },
                root_map={
                    'validations': {
                        ('limit',): {

                            'inclusive_maximum': 500,
                            'inclusive_minimum': 1,
                        },
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'limit':
                            (int,),
                        'name':
                            (str,),
                        'page_at_id':
                            (str,),
                        'org_id':
                            (str,),
                        'label':
                            (str,),
                    },
                    'attribute_map': {
                        'limit': 'limit',
                        'name': 'name',
                        'page_at_id': 'page_at_id',
                        'org_id': 'org_id',
                        'label': 'label',
                    },
                    'location_map': {
                        'limit': 'query',
                        'name': 'query',
                        'page_at_id': 'query',
                        'org_id': 'query',
                        'label': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__list_standalone_rulesets
            )

        def __replace_standalone_rule(
            self,
            rule_id,
            **kwargs
        ):
            """update a standalone rule  # noqa: E501

            update a standalone rule  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.replace_standalone_rule(rule_id, async_req=True)
            >>> result = thread.get()

            Args:
                rule_id (str): The id of a rule

            Keyword Args:
                standalone_rule (StandaloneRule): [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                StandaloneRule
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['rule_id'] = \
                rule_id
            return self.call_with_http_info(**kwargs)

        if self.replace_standalone_rule is None:
            self.replace_standalone_rule = _Endpoint(
                settings={
                    'response_type': (StandaloneRule,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_rules/{rule_id}',
                    'operation_id': 'replace_standalone_rule',
                    'http_method': 'PUT',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'rule_id',
                        'standalone_rule',
                    ],
                    'required': [
                        'rule_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'rule_id':
                            (str,),
                        'standalone_rule':
                            (StandaloneRule,),
                    },
                    'attribute_map': {
                        'rule_id': 'rule_id',
                    },
                    'location_map': {
                        'rule_id': 'path',
                        'standalone_rule': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__replace_standalone_rule
            )

        def __replace_standalone_rule_tree(
            self,
            standalone_rule_tree_id,
            **kwargs
        ):
            """update a standalone rule tree  # noqa: E501

            update a standalone rule tree  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.replace_standalone_rule_tree(standalone_rule_tree_id, async_req=True)
            >>> result = thread.get()

            Args:
                standalone_rule_tree_id (str): The id of a rule tree

            Keyword Args:
                standalone_rule_tree (StandaloneRuleTree): [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                StandaloneRuleTree
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['standalone_rule_tree_id'] = \
                standalone_rule_tree_id
            return self.call_with_http_info(**kwargs)

        if self.replace_standalone_rule_tree is None:
            self.replace_standalone_rule_tree = _Endpoint(
                settings={
                    'response_type': (StandaloneRuleTree,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_rule_trees/{standalone_rule_tree_id}',
                    'operation_id': 'replace_standalone_rule_tree',
                    'http_method': 'PUT',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'standalone_rule_tree_id',
                        'standalone_rule_tree',
                    ],
                    'required': [
                        'standalone_rule_tree_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'standalone_rule_tree_id':
                            (str,),
                        'standalone_rule_tree':
                            (StandaloneRuleTree,),
                    },
                    'attribute_map': {
                        'standalone_rule_tree_id': 'standalone_rule_tree_id',
                    },
                    'location_map': {
                        'standalone_rule_tree_id': 'path',
                        'standalone_rule_tree': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__replace_standalone_rule_tree
            )

        def __replace_standalone_ruleset(
            self,
            standalone_ruleset_id,
            **kwargs
        ):
            """update a standalone ruleset  # noqa: E501

            update a standalone ruleset  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.replace_standalone_ruleset(standalone_ruleset_id, async_req=True)
            >>> result = thread.get()

            Args:
                standalone_ruleset_id (str): The id of a ruleset

            Keyword Args:
                standalone_ruleset (StandaloneRuleset): [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                StandaloneRuleset
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['standalone_ruleset_id'] = \
                standalone_ruleset_id
            return self.call_with_http_info(**kwargs)

        if self.replace_standalone_ruleset is None:
            self.replace_standalone_ruleset = _Endpoint(
                settings={
                    'response_type': (StandaloneRuleset,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_rulesets/{standalone_ruleset_id}',
                    'operation_id': 'replace_standalone_ruleset',
                    'http_method': 'PUT',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'standalone_ruleset_id',
                        'standalone_ruleset',
                    ],
                    'required': [
                        'standalone_ruleset_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'standalone_ruleset_id':
                            (str,),
                        'standalone_ruleset':
                            (StandaloneRuleset,),
                    },
                    'attribute_map': {
                        'standalone_ruleset_id': 'standalone_ruleset_id',
                    },
                    'location_map': {
                        'standalone_ruleset_id': 'path',
                        'standalone_ruleset': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__replace_standalone_ruleset
            )

        def __replace_standalone_ruleset_bundle(
            self,
            standalone_ruleset_bundle_id,
            **kwargs
        ):
            """update a standalone ruleset bundle  # noqa: E501

            update a standalone ruleset bundle  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.replace_standalone_ruleset_bundle(standalone_ruleset_bundle_id, async_req=True)
            >>> result = thread.get()

            Args:
                standalone_ruleset_bundle_id (str): The id of a ruleset

            Keyword Args:
                standalone_ruleset_bundle (StandaloneRulesetBundle): [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                StandaloneRulesetBundle
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['standalone_ruleset_bundle_id'] = \
                standalone_ruleset_bundle_id
            return self.call_with_http_info(**kwargs)

        if self.replace_standalone_ruleset_bundle is None:
            self.replace_standalone_ruleset_bundle = _Endpoint(
                settings={
                    'response_type': (StandaloneRulesetBundle,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/standalone_ruleset_bundles/{standalone_ruleset_bundle_id}',
                    'operation_id': 'replace_standalone_ruleset_bundle',
                    'http_method': 'PUT',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'standalone_ruleset_bundle_id',
                        'standalone_ruleset_bundle',
                    ],
                    'required': [
                        'standalone_ruleset_bundle_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'standalone_ruleset_bundle_id':
                            (str,),
                        'standalone_ruleset_bundle':
                            (StandaloneRulesetBundle,),
                    },
                    'attribute_map': {
                        'standalone_ruleset_bundle_id': 'standalone_ruleset_bundle_id',
                    },
                    'location_map': {
                        'standalone_ruleset_bundle_id': 'path',
                        'standalone_ruleset_bundle': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__replace_standalone_ruleset_bundle
            )

    create_ruleset_label = None 
    create_standalone_rule = None 
    create_standalone_rule_tree = None 
    create_standalone_ruleset = None 
    create_standalone_ruleset_bundle = None 
    delete_ruleset_label = None 
    delete_standalone_rule = None 
    delete_standalone_rule_tree = None 
    delete_standalone_ruleset = None 
    delete_standalone_ruleset_bundle = None 
    get_ruleset_label = None 
    get_standalone_rule = None 
    get_standalone_rule_tree = None 
    get_standalone_ruleset = None 
    get_standalone_ruleset_bundle = None 
    list_ruleset_labels = None 
    list_standalone_rule_trees = None 
    list_standalone_rules = None 
    list_standalone_ruleset_bundles = None 
    list_standalone_rulesets = None 
    replace_standalone_rule = None 
    replace_standalone_rule_tree = None 
    replace_standalone_ruleset = None 
    replace_standalone_ruleset_bundle = None 



import datetime
import os
import agilicus
from agilicus import ApiException


import json
from .input_helpers import get_org_from_input_or_ctx
from .input_helpers import pop_item_if_none
from .output import output_if_console
from .context import get_apiclient_from_ctx
import operator

from .output.table import (
    column,
    spec_column,
    status_column,
    metadata_column,
    format_table,
    subtable,
    format_timestamp,
    format_currency_from_cents,
    format_date_only,
)


def delete_billing_account(ctx, billing_account_id=None, **kwargs):
    client = get_apiclient_from_ctx(ctx)
    return client.billing_api.delete_billing_account(billing_account_id, **kwargs)


def delete_subscription(ctx, billing_subscription_id, **kwargs):
    client = get_apiclient_from_ctx(ctx)
    return client.billing_api.delete_subscription(billing_subscription_id, **kwargs)


def get_billing_account(ctx, billing_account_id=None, **kwargs):
    client = get_apiclient_from_ctx(ctx)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    if org_id:
        kwargs["org_id"] = org_id
    else:
        kwargs.pop("org_id")
    return client.billing_api.get_billing_account(billing_account_id, **kwargs)


def update_subscription(ctx, billing_subscription_id, subscription, **kwargs):
    client = get_apiclient_from_ctx(ctx)
    return client.billing_api.replace_subscription(
        billing_subscription_id, billing_org_subscription=subscription, **kwargs
    )


def get_billing_subscription(ctx, billing_subscription_id, **kwargs):
    client = get_apiclient_from_ctx(ctx)
    return client.billing_api.get_subscription(billing_subscription_id, **kwargs)


def list_accounts(ctx, **kwargs):
    client = get_apiclient_from_ctx(ctx)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    if org_id:
        kwargs["org_id"] = org_id
    else:
        kwargs.pop("org_id")
    return client.billing_api.list_billing_accounts(**kwargs)


def list_subscriptions(ctx, **kwargs):
    client = get_apiclient_from_ctx(ctx)
    pop_item_if_none(kwargs)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    if org_id:
        kwargs["org_id"] = org_id
    else:
        kwargs.pop("org_id")
    return client.billing_api.list_subscriptions(**kwargs)


def format_accounts(
    ctx, accounts, get_subscription_data=False, get_customer_data=False, **kwargs
):
    orgs_column = [column("id"), column("organisation")]
    subscriptions = [
        metadata_column("id"),
        spec_column("subscription_id", out_name="stripe subscription"),
        subtable(ctx, "orgs", orgs_column, subobject_name="status"),
    ]
    products_column = [
        column("name", optional=True),
    ]
    columns = [
        metadata_column("id"),
        status_column("product.spec.name", optional=True),
    ]

    def _get_customer_name(record, key):
        status = record.get("status")
        return status.get("customer", {}).get("name")

    if get_customer_data:
        columns.append(
            column(
                "status", newname="customer", getter=_get_customer_name, optional=True
            )
        )
    else:
        columns.append(spec_column("customer_id"))

    if get_subscription_data:
        columns.append(
            subtable(ctx, "products", products_column, subobject_name="status")
        )

    columns.append(subtable(ctx, "orgs", orgs_column, subobject_name="status"))
    columns.append(
        subtable(ctx, "org_subscriptions", subscriptions, subobject_name="status")
    )
    return format_table(ctx, accounts, columns)


def format_subscriptions(
    ctx, subscriptions, get_usage_metrics=False, get_stripe_status=False, **kwargs
):
    orgs_column = [column("id"), column("organisation")]

    def _convert_amount(record, amount):
        status = record.get("status")
        subscription = status.get("subscription")
        currency = subscription.get("currency")
        return format_currency_from_cents(amount, currency)

    def _get_amount(record, key):
        status = record.get("status")
        balance = status.get("balance")
        amount = balance.get("subscription_balance")
        return _convert_amount(record, amount)

    def _get_invoice_total(record, key):
        status = record.get("status")
        invoice = status.get("balance", {}).get("upcoming_invoice")
        return _convert_amount(record, invoice.get("total"))

    columns = [
        metadata_column("id", optional=True),
        spec_column("subscription_id"),
        spec_column("product_id"),
        subtable(ctx, "orgs", orgs_column, subobject_name="status"),
        column(
            "invoice.total",
            newname="upcoming invoice",
            getter=_get_invoice_total,
            optional=True,
        ),
        column("subscription_balance", optional=True, getter=_get_amount),
        column(
            "status.balance.estimate_balance_end_date",
            newname="end date estimate",
            format_fn=format_date_only,
            optional=True,
        ),
    ]

    if get_usage_metrics:
        metrics_columns = [column("type"), column("active")]
        columns.append(
            subtable(
                ctx, "metrics", metrics_columns, subobject_name="status.usage_metrics"
            )
        )

    if get_stripe_status:
        columns.append(status_column("provider_status"))

    return format_table(ctx, subscriptions, columns)


def add_billing_account(ctx, customer_id=None, dev_mode=None, **kwargs):
    client = get_apiclient_from_ctx(ctx)
    spec = agilicus.BillingAccountSpec(customer_id=customer_id)

    if dev_mode is not None:
        spec.dev_mode = dev_mode

    account = agilicus.BillingAccount(spec=spec)

    return client.billing_api.create_billing_account(account)


def add_subscription_balance_transaction(
    ctx, billing_subscription_id, amount=None, adjustment_type="credit", **kwargs
):
    client = get_apiclient_from_ctx(ctx)
    transaction = agilicus.BillingBalanceTransaction(**kwargs)
    if amount is not None:
        # convert to cents
        transaction.amount = int(amount * 100)
    if adjustment_type == "credit":
        # debits are 'negative', are removed.
        transaction.amount = -transaction.amount

    return client.billing_api.add_subscription_balance_transaction(
        billing_subscription_id, billing_balance_transaction=transaction
    )


def list_subscription_balance_transactions(ctx, billing_subscription_id, **kwargs):
    client = get_apiclient_from_ctx(ctx)
    pop_item_if_none(kwargs)
    return client.billing_api.get_subscription_balance_transactions(
        billing_subscription_id, **kwargs
    )


def list_customer_balance_transactions(ctx, billing_id, **kwargs):
    client = get_apiclient_from_ctx(ctx)
    pop_item_if_none(kwargs)
    return client.billing_api.get_customer_balance_transactions(billing_id, **kwargs)


def format_balance_transactions(ctx, results):
    def get_amount(record, key):
        amount = record.get("amount")
        currency = record.get("currency")
        return format_currency_from_cents(amount, currency)

    columns = [
        column("id"),
        column("created", format_fn=format_timestamp),
        column("description"),
        column("amount", getter=get_amount),
        column("currency"),
        column("type"),
        # column("invoice"),
    ]
    return format_table(ctx, results, columns)


def add_subscription(
    ctx, billing_account_id, subscription_id=None, dev_mode=None, **kwargs
):
    client = get_apiclient_from_ctx(ctx)
    spec = agilicus.BillingOrgSubscriptionSpec(billing_account_id=billing_account_id)

    if dev_mode is not None:
        spec.dev_mode = dev_mode

    if subscription_id is not None:
        spec.subscription_id = subscription_id

    subscription = agilicus.BillingOrgSubscription(spec=spec)

    return client.billing_api.create_subscription(subscription)


def add_org(ctx, billing_account_id=None, org_id=None, **kwargs):
    client = get_apiclient_from_ctx(ctx)
    billing_org = agilicus.BillingOrg._from_openapi_data(org_id=org_id)
    return client.billing_api.add_org_to_billing_account(
        billing_account_id, billing_org=billing_org
    )


def remove_org(ctx, billing_account_id=None, org_id=None, **kwargs):
    client = get_apiclient_from_ctx(ctx)
    return client.billing_api.remove_org_from_billing_account(billing_account_id, org_id)


def add_org_to_subscription(ctx, billing_subscription_id, org_id, **kwargs):
    client = get_apiclient_from_ctx(ctx)
    billing_org = agilicus.BillingOrg._from_openapi_data(org_id=org_id)
    return client.billing_api.add_org_to_billing_subscription(
        billing_subscription_id, billing_org=billing_org
    )


def remove_org_from_subscription(ctx, billing_subscription_id, org_id, **kwargs):
    client = get_apiclient_from_ctx(ctx)
    return client.billing_api.remove_org_from_billing_subscription(
        org_id, billing_subscription_id
    )


def replace_billing_account(
    ctx,
    billing_account_id=None,
    customer_id=None,
    dev_mode=None,
    product_id=None,
    **kwargs,
):
    client = get_apiclient_from_ctx(ctx)

    existing = client.billing_api.get_billing_account(billing_account_id)
    if customer_id is not None:
        existing.spec.customer_id = customer_id
    if dev_mode is not None:
        existing.spec.dev_mode = dev_mode
    if product_id is not None:
        existing.spec.product_id = product_id
    return client.billing_api.replace_billing_account(
        billing_account_id, billing_account=existing
    )


def format_usage_records(ctx, records):
    columns = [
        column("id"),
        column("period"),
        column("total_usage"),
    ]
    return format_table(ctx, records, columns, getter=operator.itemgetter)


def get_usage_records(ctx, billing_account_id=None, **kwargs):
    client = get_apiclient_from_ctx(ctx)
    return client.billing_api.get_usage_records(billing_account_id)


def _dump_billing_failure(ctx, exception_description, account):
    account_info = {}
    try:
        account_info = account.to_dict()
    except Exception as exc:
        output_if_console(
            ctx,
            f"Failed to dump account info when handling billing failure: {str(exc)}",
        )
    error_message = {
        "time": datetime.datetime.now(datetime.timezone.utc),
        "msg": "error",
        "reason": str(exception_description),
        "account": account_info,
    }
    try:
        print(json.dumps(error_message, default=str))
    except Exception as exc:
        output_if_console(ctx, f"Failed to json.dumps failure info: {str(exc)}")


def run_billing_um_all_accounts(
    ctx, client, dry_run=False, push_to_prometheus_on_success=True, **kwargs
):
    accounts = client.billing_api.list_billing_accounts()
    record = agilicus.CreateBillingUsageRecords(dry_run=dry_run)
    numSuccess = 0
    numSkipped = 0
    numFail = 0

    for account in accounts.billing_accounts:

        if not account.spec.customer_id:
            numSkipped += 1
            continue
        try:

            if len(account.status.orgs) == 0:
                numSkipped += 1
                print(
                    json.dumps(
                        {
                            "skip": True,
                            "billing_account": account.metadata.id,
                            "customer_id": account.spec.customer_id,
                        }
                    )
                )
                continue

            base_result = client.billing_api.add_billing_usage_record(
                account.metadata.id, create_billing_usage_records=record
            )
            success = False
            if base_result:
                result = base_result.to_dict()
                success = True
            else:
                result = {}

            result["billing_account"] = account.metadata.id
            result["customer_id"] = account.spec.customer_id
            result["orgs"] = [
                {"id": org.id, "organisation": org.organisation}
                for org in account.status.orgs
            ]
            if success:
                numSuccess += 1
                result["published"] = True
            else:
                numSkipped += 1
                result["published"] = False
            print(json.dumps(result))
        except ApiException as exc:
            numFail += 1
            _dump_billing_failure(ctx, exc.body, account)
        except Exception as exc:
            numFail += 1
            _dump_billing_failure(ctx, exc, account)

    if push_to_prometheus_on_success:
        try:
            from prometheus_client import (
                CollectorRegistry,
                Gauge,
                push_to_gateway,
            )
        except ModuleNotFoundError:
            output_if_console(ctx, "Not posting success to prometheus_client.")
            output_if_console(
                ctx, "Add the 'billing' option to the install to gain access"
            )
            return

        registry = CollectorRegistry()
        gSuccess = Gauge(
            "billing_usage_records_created_count",
            "number of billing accounts that have created a usage record",
            registry=registry,
        )

        gFail = Gauge(
            "billing_usage_records_failed_count",
            "number of billing accounts that failed to create a usage record",
            registry=registry,
        )
        gSkipped = Gauge(
            "billing_usage_records_skipped_count",
            "number of billing accounts that were skipped",
            registry=registry,
        )

        push_gateway = os.environ.get(
            "PROMETHEUS_PUSH_GATEWAY",
            "push-prometheus-pushgateway.prometheus-pushgateway:9091",
        )
        job_name = os.environ.get("JOB_NAME", "billing_usage_job")
        gSuccess.set(numSuccess)
        gFail.set(numFail)
        gSkipped.set(numSkipped)
        push_to_gateway(push_gateway, job=job_name, registry=registry)


def create_usage_record(
    ctx, billing_account_id=None, all_accounts=None, dry_run=False, **kwargs
):
    client = get_apiclient_from_ctx(ctx)
    record = agilicus.BillingUsageRecord(dry_run=dry_run)
    if billing_account_id is not None:
        records = agilicus.CreateBillingUsageRecords(usage_records=[record])
        result = client.billing_api.add_billing_usage_record(
            billing_account_id, create_billing_usage_records=records
        )
        print(json.dumps(result.to_dict()))
    elif all_accounts is not None:
        run_billing_um_all_accounts(ctx, client, dry_run=dry_run, **kwargs)
    else:
        raise Exception("Need to choose --billing-account-or or --all-accounts")


def _get_subscription(org_subscriptions, subscription_id):
    for org_subscription in org_subscriptions:
        if org_subscription.spec.subscription_id == subscription_id:
            return org_subscription


def _has_org(orgs, org_id):
    for org in orgs:
        if org.id == org_id:
            return True
    return False


def migrate_billing_subscriptions(ctx, billing_account_id=None, commit=False, **kwargs):
    kwargs = {}
    kwargs["get_customer_data"] = True
    kwargs["get_subscription_data"] = True
    kwargs["org_id"] = ""
    if billing_account_id is not None:
        accounts = [get_billing_account(ctx, billing_account_id, **kwargs)]
    else:
        accounts = list_accounts(ctx, **kwargs).billing_accounts
    for account in accounts:
        customer = account.status.customer
        if not customer:
            print(f"billing account {account.metadata.id} has no customer")
            continue
        if customer.get("deleted"):
            continue
        for sub in account.status.subscriptions:
            # check if there is an org sub for this first.
            org_subscription = _get_subscription(
                account.status.org_subscriptions, sub["id"]
            )
            if not org_subscription:
                print(
                    f"  billing account:{account.metadata.id} "
                    "{customer['name']}:{sub['id']} "
                    "needs BillingSubscription"
                )
                if not commit:
                    for org in account.status.orgs:
                        print(f"    {org.organisation} needs to be attached")
                    continue
                sub = add_subscription(
                    ctx,
                    account.metadata.id,
                    subscription_id=sub["id"],
                    dev_mode=account.spec.dev_mode,
                )
                print(f"    created billing subscription {sub.metadata.id}")
                for org in account.status.orgs:
                    print(f"    adding org {org.organisation} to billing subscription")
                    val = add_org_to_subscription(ctx, sub.metadata.id, org.id)
                    print(f"    added org {val.id} to billing subscription")
            else:
                # verify that the orgs are added to subscription
                for org in account.status.orgs:
                    if not _has_org(org_subscription.status.orgs, org.id):
                        if org.admin_state != "active":
                            # not interested in orgs that are not active
                            continue
                        print(
                            f"    NOTICE: org {org.organisation} is missing billing "
                            f"in subscription {org_subscription.metadata.id}",
                        )
                        if not commit:
                            continue
                        val = add_org_to_subscription(
                            ctx, org_subscription.metadata.id, org.id
                        )
                        print(f"    added org {val.id} to billing subscription")


def reconcile_billing_subscriptions(
    ctx, billing_account_id=None, commit=False, push_to_prometheus=False, **kwargs
):
    """
    find cancelled stripe subscriptions with an active organsanistion(s).
    """
    kwargs = {}
    kwargs["get_customer_data"] = True
    kwargs["get_subscription_data"] = True
    kwargs["org_id"] = ""
    subscriptionsGood = 0
    subscriptionsNotActive = 0
    subscriptionOrgNotActive = 0
    if billing_account_id is not None:
        accounts = [get_billing_account(ctx, billing_account_id, **kwargs)]
    else:
        accounts = list_accounts(ctx, **kwargs).billing_accounts
    results = []
    for account in accounts:
        customer = account.status.customer
        if not customer:
            print(f"billing account {account.metadata.id} has no customer")
            continue
        for sub in account.status.subscriptions:
            counts = {}
            subscriptionGood = True
            for org in account.status.orgs:
                admin_state = org.admin_state.value
                val = counts.setdefault(admin_state, 0)
                counts[admin_state] = val + 1
                if (
                    admin_state == "active"
                    and sub["status"] != "active"
                    and sub["status"] != "trialing"
                ):
                    print(
                        "  NOTICE: Subscription cancelled, "
                        f"but org {org.id} ({org.organisation}) still active "
                    )
                    results.append({"org": org, "sub": sub})
                    subscriptionsNotActive += 1
                    subscriptionGood = False
            if counts.get("active", 0) == 0 and (
                sub["status"] == "active" or sub["status"] == "trialing"
            ):
                print(
                    "  NOTICE: All orgs deleted in billing account "
                    f"{account.metadata.id}, subscription still active "
                    f"{sub['status']}"
                )
                results.append({"org": org, "sub": sub})
                subscriptionGood = False
                subscriptionOrgNotActive += 1
            if subscriptionGood:
                subscriptionsGood += 1

    def org_getter(record, key):
        return record["org"]["organisation"]

    def org_id_getter(record, key):
        return record["org"]["id"]

    def state_getter(record, key):
        return record["org"]["admin_state"]

    def sub_getter(record, key):
        return record["sub"]["status"]

    columns = [
        column("organisation", getter=org_getter),
        column("id", getter=org_id_getter),
        column("admin_state", getter=state_getter),
        column("stripe subscription status", getter=sub_getter),
    ]
    print(format_table(ctx, results, columns))

    if push_to_prometheus:
        try:
            from prometheus_client import (
                CollectorRegistry,
                Gauge,
                push_to_gateway,
            )
        except ModuleNotFoundError:
            output_if_console(ctx, "Not posting success to prometheus_client.")
            output_if_console(
                ctx, "Add the 'billing' option to the install to gain access"
            )
            return

        registry = CollectorRegistry()
        gSubscriptionsGood = Gauge(
            "billing_subscriptions_good",
            "billing accounts that have an active org and active subscription",
            registry=registry,
        )
        gSubscriptionsNotActive = Gauge(
            "billing_subscriptions_not_active",
            "billing accounts that have an active org but inactive subscription",
            registry=registry,
        )
        gOrgNotActive = Gauge(
            "billing_subscriptions_organisation_not_active",
            "billing accounts that have disable org but active subscription",
            registry=registry,
        )
        gSubscriptionsGood.set(subscriptionsGood)
        gSubscriptionsNotActive.set(subscriptionsNotActive)
        gOrgNotActive.set(subscriptionOrgNotActive)

        push_gateway = os.environ.get(
            "PROMETHEUS_PUSH_GATEWAY",
            "push-prometheus-pushgateway.prometheus-pushgateway:9091",
        )
        job_name = os.environ.get("JOB_NAME", "billing_reconcile_job")
        push_to_gateway(push_gateway, job=job_name, registry=registry)


def list_subscription_features(ctx, billing_subscription_id, **kwargs):
    client = get_apiclient_from_ctx(ctx)
    return client.billing_api.list_subscription_features(
        billing_subscription_id, **kwargs
    ).features



"""
    Agilicus API

    Agilicus is API-first. Modern software is controlled by other software, is open, is available for you to use the way you want, securely, simply.  The OpenAPI Specification in YAML format is available on [www](https://www.agilicus.com/www/api/agilicus-openapi.yaml) for importing to other tools.  A rendered, online viewable and usable version of this specification is available at [api](https://www.agilicus.com/api). You may try the API inline directly in the web page. To do so, first obtain an Authentication Token (the simplest way is to install the Python SDK, and then run `agilicus-cli --issuer https://MYISSUER get-token`). You will need an org-id for most calls (and can obtain from `agilicus-cli --issuer https://MYISSUER list-orgs`). The `MYISSUER` will typically be `auth.MYDOMAIN`, and you will see it as you sign-in to the administrative UI.  This API releases on Bearer-Token authentication. To obtain a valid bearer token you will need to Authenticate to an Issuer with OpenID Connect (a superset of OAUTH2).  Your \"issuer\" will look like https://auth.MYDOMAIN. For example, when you signed-up, if you said \"use my own domain name\" and assigned a CNAME of cloud.example.com, then your issuer would be https://auth.cloud.example.com.  If you selected \"use an Agilicus supplied domain name\", your issuer would look like https://auth.myorg.agilicus.cloud.  For test purposes you can use our [Python SDK](https://pypi.org/project/agilicus/) and run `agilicus-cli --issuer https://auth.MYDOMAIN get-token`.  This API may be used in any language runtime that supports OpenAPI 3.0, or, you may use our [Python SDK](https://pypi.org/project/agilicus/), our [Typescript SDK](https://www.npmjs.com/package/@agilicus/angular), or our [Golang SDK](https://git.agilicus.com/pub/sdk-go).  100% of the activities in our system our API-driven, from our web-admin, through our progressive web applications, to all internals: there is nothing that is not accessible.  For more information, see [developer resources](https://www.agilicus.com/developer).   # noqa: E501

    The version of the OpenAPI document: 2024.01.29
    Contact: dev@agilicus.com
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from agilicus_api.api_client import ApiClient, Endpoint as _Endpoint
from agilicus_api.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types
)
from agilicus_api.model.list_logs_response import ListLogsResponse


class DiagnosticsApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        def __list_logs(
            self,
            org_id,
            **kwargs
        ):
            """Retrieve application logs  # noqa: E501

            Retrieve application diagnostic logs. These are the output from stdout/stderr (stream).   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.list_logs(org_id, async_req=True)
            >>> result = thread.get()

            Args:
                org_id (str): query for a specific organization

            Keyword Args:
                dt_from (str): search criteria, search logs from. [optional]
                dt_to (str): search criteria, search logs to. [optional]
                app (str): search criteria, search logs for an app. [optional]
                sub (str): search criteria, search logs for a given subject (USER) GUID. [optional]
                limit (int): limit number of output logs. [optional] if omitted the server will use the default value of 150
                sub_org_id (str): query for a specific sub-organization. [optional]
                env (str): query for a specific environment. [optional]
                dt_sort (str): Sort order: *'asc' - ascending, from old to new logs *'desc' - descending, from new to old logs . [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ListLogsResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['org_id'] = \
                org_id
            return self.call_with_http_info(**kwargs)

        if self.list_logs is None:
            self.list_logs = _Endpoint(
                settings={
                    'response_type': (ListLogsResponse,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/diagnostics/logs',
                    'operation_id': 'list_logs',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'org_id',
                        'dt_from',
                        'dt_to',
                        'app',
                        'sub',
                        'limit',
                        'sub_org_id',
                        'env',
                        'dt_sort',
                    ],
                    'required': [
                        'org_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                        'dt_sort',
                    ],
                    'validation': [
                        'limit',
                    ]
                },
                root_map={
                    'validations': {
                        ('limit',): {

                            'inclusive_maximum': 500,
                            'inclusive_minimum': 1,
                        },
                    },
                    'allowed_values': {
                        ('dt_sort',): {

                            "ASC": "asc",
                            "DESC": "desc"
                        },
                    },
                    'openapi_types': {
                        'org_id':
                            (str,),
                        'dt_from':
                            (str,),
                        'dt_to':
                            (str,),
                        'app':
                            (str,),
                        'sub':
                            (str,),
                        'limit':
                            (int,),
                        'sub_org_id':
                            (str,),
                        'env':
                            (str,),
                        'dt_sort':
                            (str,),
                    },
                    'attribute_map': {
                        'org_id': 'org_id',
                        'dt_from': 'dt_from',
                        'dt_to': 'dt_to',
                        'app': 'app',
                        'sub': 'sub',
                        'limit': 'limit',
                        'sub_org_id': 'sub_org_id',
                        'env': 'env',
                        'dt_sort': 'dt_sort',
                    },
                    'location_map': {
                        'org_id': 'query',
                        'dt_from': 'query',
                        'dt_to': 'query',
                        'app': 'query',
                        'sub': 'query',
                        'limit': 'query',
                        'sub_org_id': 'query',
                        'env': 'query',
                        'dt_sort': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__list_logs
            )

    list_logs = None 



from .. import context
import json
from ..input_helpers import get_org_from_input_or_ctx, strip_none
import agilicus
from ..output.table import (
    spec_column,
    format_table,
    metadata_column,
    subtable,
    column,
)


def add_host(ctx, hostname, labels, port=None, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    kwargs = strip_none(kwargs)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id
    spec = agilicus.HostSpec(agilicus.Domain(hostname), **kwargs)
    if port:
        spec.port = agilicus.NetworkPort(port)

    if labels:
        spec.labels = [agilicus.HostLabelName(label) for label in labels]
    req = agilicus.Host(spec=spec)
    try:
        return apiclient.hosts_api.add_host(req)
    except Exception as exc:
        if exc.status == 409:
            body = exc.body
            if not body:
                raise
            result = json.loads(body)
            guid = agilicus.find_guid(result)
            req = apiclient.hosts_api.get_host(guid, org_id=org_id)
            if not req.spec.labels:
                req.spec.labels = []
            for label in labels:
                req.spec.labels.append(agilicus.TrustedCertificateLabelName(label))
            return apiclient.hosts_api.replace_host(
                req.metadata.id,
                req,
            )
        else:
            raise


def get_host(ctx, host_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id

    kwargs = strip_none(kwargs)
    return apiclient.hosts_api.get_host(
        host_id,
        **kwargs,
    )


def delete_host(ctx, host_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id

    kwargs = strip_none(kwargs)
    return apiclient.hosts_api.delete_host(
        host_id,
        **kwargs,
    )


def add_label(ctx, label, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id

    labelName = agilicus.HostLabelName(label)
    spec = agilicus.HostLabelSpec(labelName, **kwargs)
    req = agilicus.HostLabel(spec=spec)
    return apiclient.hosts_api.add_host_label(req)


def add_bundle(ctx, bundle, label=None, label_org_id=None, exclude=None, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id
    if label_org_id is None:
        label_org_id = org_id
    bundle_label = None
    if label is not None:
        label_name = agilicus.HostLabelName(label)
        label_spec = agilicus.HostLabelSpec(name=label_name, org_id=label_org_id)
        bundle_label = agilicus.HostBundleLabel(label=label_spec)
        if exclude is not None:
            bundle_label.exclude = exclude

    bundle_name = agilicus.HostBundleName(bundle)
    spec = agilicus.HostBundleSpec(bundle_name, **kwargs)
    if bundle_label:
        spec.labels = []
        spec.labels.append(bundle_label)
    req = agilicus.HostBundle(spec=spec)
    return apiclient.hosts_api.add_host_bundle(req)


def add_bundle_label(ctx, bundle_id, label, label_org_id=None, exclude=None, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id
    if label_org_id is None:
        label_org_id = org_id

    bundle = get_bundle(ctx, bundle_id, **kwargs)
    bundle_label = None
    label_name = agilicus.HostLabelName(label)
    label_spec = agilicus.HostLabelSpec(name=label_name, org_id=label_org_id)
    bundle_label = agilicus.HostBundleLabel(label=label_spec)
    if exclude is not None:
        bundle_label.exclude = exclude

    if not bundle.spec.labels:
        bundle.spec.labels = []

    bundle.spec.labels.append(bundle_label)
    return apiclient.hosts_api.replace_host_bundle(bundle_id, bundle)


def delete_bundle(ctx, bundle_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id
    return apiclient.hosts_api.delete_host_bundle(bundle_id, **kwargs)


def list_bundles(ctx, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id

    kwargs = strip_none(kwargs)
    return apiclient.hosts_api.list_host_bundles(
        **kwargs,
    ).host_bundles


def get_bundle(ctx, bundle_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id

    kwargs = strip_none(kwargs)
    return apiclient.hosts_api.get_host_bundle(
        bundle_id,
        **kwargs,
    )


def format_bundles(ctx, bundles):
    label_columns = [
        column("label"),
    ]
    columns = [
        metadata_column("id"),
        metadata_column("created"),
        spec_column("org_id"),
        spec_column("name"),
        subtable(ctx, "spec.labels", label_columns),
    ]
    return format_table(ctx, bundles, columns)


def list_labels(ctx, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id

    kwargs = strip_none(kwargs)
    return apiclient.hosts_api.list_host_labels(
        **kwargs,
    ).host_labels


def list_hosts(ctx, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)

    kwargs = strip_none(kwargs)
    return apiclient.hosts_api.list_hosts(org_id=org_id, **kwargs).hosts


def format_labels(ctx, labels):
    columns = [
        metadata_column("created"),
        spec_column("org_id"),
        spec_column("name"),
    ]
    return format_table(ctx, labels, columns)


def format_hosts_as_text(ctx, hosts):
    columns = [
        metadata_column("id"),
        metadata_column("created"),
        spec_column("org_id"),
        spec_column("hostname"),
        spec_column("path"),
        spec_column("port"),
        spec_column("labels"),
    ]
    return format_table(ctx, hosts, columns)


def list_orgs(ctx, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id

    kwargs = strip_none(kwargs)
    return apiclient.hosts_api.list_host_orgs(
        **kwargs,
    ).orgs


def format_orgs(ctx, orgs):
    columns = [
        column("id"),
        column("organisation"),
    ]
    return format_table(ctx, orgs, columns)



"""
    Agilicus API

    Agilicus is API-first. Modern software is controlled by other software, is open, is available for you to use the way you want, securely, simply.  The OpenAPI Specification in YAML format is available on [www](https://www.agilicus.com/www/api/agilicus-openapi.yaml) for importing to other tools.  A rendered, online viewable and usable version of this specification is available at [api](https://www.agilicus.com/api). You may try the API inline directly in the web page. To do so, first obtain an Authentication Token (the simplest way is to install the Python SDK, and then run `agilicus-cli --issuer https://MYISSUER get-token`). You will need an org-id for most calls (and can obtain from `agilicus-cli --issuer https://MYISSUER list-orgs`). The `MYISSUER` will typically be `auth.MYDOMAIN`, and you will see it as you sign-in to the administrative UI.  This API releases on Bearer-Token authentication. To obtain a valid bearer token you will need to Authenticate to an Issuer with OpenID Connect (a superset of OAUTH2).  Your \"issuer\" will look like https://auth.MYDOMAIN. For example, when you signed-up, if you said \"use my own domain name\" and assigned a CNAME of cloud.example.com, then your issuer would be https://auth.cloud.example.com.  If you selected \"use an Agilicus supplied domain name\", your issuer would look like https://auth.myorg.agilicus.cloud.  For test purposes you can use our [Python SDK](https://pypi.org/project/agilicus/) and run `agilicus-cli --issuer https://auth.MYDOMAIN get-token`.  This API may be used in any language runtime that supports OpenAPI 3.0, or, you may use our [Python SDK](https://pypi.org/project/agilicus/), our [Typescript SDK](https://www.npmjs.com/package/@agilicus/angular), or our [Golang SDK](https://git.agilicus.com/pub/sdk-go).  100% of the activities in our system our API-driven, from our web-admin, through our progressive web applications, to all internals: there is nothing that is not accessible.  For more information, see [developer resources](https://www.agilicus.com/developer).   # noqa: E501

    The version of the OpenAPI document: 2024.01.29
    Contact: dev@agilicus.com
    Generated by: https://openapi-generator.tech
"""


import unittest

import agilicus_api
from agilicus_api.api.organisations_api import OrganisationsApi  # noqa: E501


class TestOrganisationsApi(unittest.TestCase):
    """OrganisationsApi unit test stubs"""

    def setUp(self):
        self.api = OrganisationsApi()  # noqa: E501

    def tearDown(self):
        pass

    def test_create_billing_portal_link(self):
        """Test case for create_billing_portal_link

        Create a link to the billing portal  # noqa: E501
        """
        pass

    def test_create_blocking_upgrade_orgs_task(self):
        """Test case for create_blocking_upgrade_orgs_task

        utility to upgrade organisations  # noqa: E501
        """
        pass

    def test_create_org(self):
        """Test case for create_org

        Create an organisation  # noqa: E501
        """
        pass

    def test_create_sub_org(self):
        """Test case for create_sub_org

        Create a sub organisation  # noqa: E501
        """
        pass

    def test_delete_sub_org(self):
        """Test case for delete_sub_org

        Delete a sub organisation  # noqa: E501
        """
        pass

    def test_get_inherent_capabilities(self):
        """Test case for get_inherent_capabilities

        Get the inherent capabilities for an org  # noqa: E501
        """
        pass

    def test_get_org(self):
        """Test case for get_org

        Get a single organisation  # noqa: E501
        """
        pass

    def test_get_org_billing_account(self):
        """Test case for get_org_billing_account

        Get the billing account associated with the organisation  # noqa: E501
        """
        pass

    def test_get_org_features(self):
        """Test case for get_org_features

        all features associated with organisation  # noqa: E501
        """
        pass

    def test_get_org_status(self):
        """Test case for get_org_status

        Get the status of an organisation  # noqa: E501
        """
        pass

    def test_get_system_options(self):
        """Test case for get_system_options

        Get organisation system options  # noqa: E501
        """
        pass

    def test_get_usage_metrics(self):
        """Test case for get_usage_metrics

        Get all usage metrics for an organisation  # noqa: E501
        """
        pass

    def test_list_email_domains(self):
        """Test case for list_email_domains

        List all unique email domains for users that are inside an organisation  # noqa: E501
        """
        pass

    def test_list_org_guid_mapping(self):
        """Test case for list_org_guid_mapping

        Get all org guids and a unique name mapping  # noqa: E501
        """
        pass

    def test_list_orgs(self):
        """Test case for list_orgs

        Get all organisations  # noqa: E501
        """
        pass

    def test_list_sub_orgs(self):
        """Test case for list_sub_orgs

        Get all sub organisations  # noqa: E501
        """
        pass

    def test_reconcile_sub_org_issuer(self):
        """Test case for reconcile_sub_org_issuer

        Creates an issuer for the sub org  # noqa: E501
        """
        pass

    def test_replace_org(self):
        """Test case for replace_org

        Create or update an organisation  # noqa: E501
        """
        pass

    def test_replace_system_options(self):
        """Test case for replace_system_options

        Update organisation system options  # noqa: E501
        """
        pass

    def test_set_inherent_capabilities(self):
        """Test case for set_inherent_capabilities

        Set the inherent capabilities for an org  # noqa: E501
        """
        pass


if __name__ == '__main__':
    unittest.main()
